#ifndef LKEYBOARD_H
#define LKEYBOARD_H

#include <LObject.h>
#include <xkbcommon/xkbcommon.h>

/**
 * @brief Class for handling keyboard events.
 *
 * The LKeyboard class lets you handle keyboard events generated by the input backend,
 * send them to clients, and configure parameters such as the keyboard map and key repeat rate.\n
 * There is a single instance of LKeyboard, which can be accessed from LSeat::keyboard().
 *
 * ### Wayland Events
 *
 * To send keyboard events to clients, you must first assign focus to a surface using setFocus().\n
 * Doing so, automatically removes focus from any previously focused surface.
 * As a result, all subsequent keyboard events sent with sendKeyEvent() or sendModifiersEvent() will be
 * directed to the currently focused surface.
 */
class Louvre::LKeyboard : public LObject
{
public:

    /**
     * @brief Constructor parameters.
     *
     * Configuration parameters passed in the LCompositor::createKeyboardRequest() virtual constructor.
     */
    struct Params;

    /**
     * @brief Keyboard modifiers.
     *
     * Stores the status of keyboard modifiers (Ctrl, Shift, Alt, etc).
     */
    struct KeyboardModifiersState
    {
        /// Active modifiers when physically pressed
        UInt32 depressed;

        /// Hooked modifiers that will be disabled after a non-modifier key is pressed
        UInt32 latched;

        /// Active modifiers until they are pressed again (e.g. the Shift key)
        UInt32 locked;

        /// Group the above states (use this value if the source of a modifier change is not of your interest)
        UInt32 group;
    };

    /**
     * @brief Key states.
     *
     * Enum with the possible states of a key.
     */
    enum KeyState : UChar8
    {
        /// The key is not being pressed
        Released = 0,

        /// The key is pressed
        Pressed = 1
    };

    /**
     * @brief Constructor of the LKeyboard class.
     *
     * There is a single instance of LKeyboard, which can be accessed from LSeat::keyboard().
     *
     * @param params Internal library parameters passed in the LCompositor::createKeyboardRequest() virtual constructor.
     */
    LKeyboard(Params *params);

    /**
     * @brief Destructor of the LKeyboard class.
     *
     * Invoked internally by the library after LCompositor::destroyKeyboardRequest() is called.
     */
    virtual ~LKeyboard();

    /// @cond OMIT
    LKeyboard(const LKeyboard&) = delete;
    LKeyboard& operator= (const LKeyboard&) = delete;
    /// @endcond

    /**
     * @brief Set keyboard focus.
     *
     * This method assigns keyboard focus to the specified surface.
     * Only one surface can have keyboard focus at a time, and this method automatically removes focus from any previously focused surface.\n
     * To remove focus from all surfaces, pass `nullptr` as the argument.
     *
     * @note Calling this method while there is a surface grabbing the keyboard is a no-op.
     */
    void setFocus(LSurface *surface);

    /**
     * @brief Surface with keyboard focus.
     *
     * Pointer to the surface with keyboard focus assigned with setFocus().
     * Only surfaces with keyboard focus can receive keyboard events.
     *
     * @returns `nullptr` if no surface has keyboard focus.
     */
    LSurface *focus() const;

    /**
     * @brief Set the surface that grabs keyboard events.
     *
     * This method redirects all keyboard events to the given surface, preventing other surfaces from gain focus with setFocus().\n
     * If the surface grabbing the keyboard is destroyed the grab is automatically assigned to its parent.\n
     * For example, an LPopupRole can request to make a keyboard grab with LPopupRole::grabSeatRequest().
     *
     * If 'surface' is set to `nullptr`, the keyboard grab is disabled.
     *
     * @param surface The surface that will grab the keyboard events or `nullptr` to disable it.
     * @param keyboardResource The specific `wl_keyboard` resource created by the client to which events must be sent.
     */
    void setGrabbingSurface(LSurface* surface, Protocols::Wayland::RKeyboard* keyboardResource);

    /**
     * @brief Get the current surface that is grabbing the keyboard events.
     *
     * This method returns the surface that is currently grabbing the keyboard events.
     * If no grab is active, it returns `nullptr`.
     *
     * @returns The surface that is grabbing keyboard events, or `nullptr` if no grab is active.
     */
    LSurface* grabbingSurface() const;

    /**
     * @brief Get the current `wl_keyboard` resource used during the keyboard grab.
     *
     * This method returns the `wl_keyboard` resource to which keyboard events are sent during the grab.
     * If no grab is active, it returns `nullptr`.
     *
     * @returns The `wl_keyboard` resource used during the keyboard grab, or `nullptr` if no grab is active.
     */
    Protocols::Wayland::RKeyboard* grabbingKeyboardResource() const;

    /**
     * @brief State of the keyboard modifiers.
     *
     * @returns An instance of KeyboardModifiersState, which stores the state of the keyboard modifiers.
     */
    const KeyboardModifiersState &modifiersState() const;

    /**
     * @brief Keyboard map file descriptor.
     *
     * File descriptor of the XKB keyboard map assigned with setKeymap().
     */
    Int32 keymapFd() const;

    /**
     * @brief Size of the keyboard map in bytes.
     *
     * Size in bytes of the XKB keyboard map loaded with setKeymap().
     */
    Int32 keymapSize() const;

    /**
     * @brief Keymap format.
     *
     * The keymap format can be either `WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1`,
     * which indicates the presence of a valid XKB keymap set, or `WL_KEYBOARD_KEYMAP_FORMAT_NO_KEYMAP`,
     * indicating an issue with the XKB system configuration.
     */
    UInt32 keymapFormat() const;

    // Since 4

    /**
     * @brief Repetition rate.
     *
     * Number of repetitions per second of a key when held down, assigned with setRepeatInfo().
     *
     * The default value is 32 repetitions per second.
     */
    Int32 repeatRate() const;

    /**
     * @brief Repetition delay.
     *
     * Delay in milliseconds before triggering the repetition of a key by holding it down, set with setRepeatInfo().
     *
     * The default value is 500 ms.
     */
    Int32 repeatDelay() const;

    /**
     * @brief Set the key repeat rate and delay.
     *
     * Assigns the repeat rate and delay when holding down a key.
     *
     * @note This impacts how clients respond to a key being held down.
     *
     * @param rate Number of repetitions per second. 0 turns off repetition.
     * @param msDelay Delay in milliseconds before triggering the replay.
     */
    void setRepeatInfo(Int32 rate, Int32 msDelay);

    /**
     * @brief Key symbol.
     *
     * Returns the symbol of a raw key code generated by the input backend according to the status of the keyboard map.\n
     *
     * The complete list of symbols can be found in [`<xkbcommon/xkbcommon-keysyms.h>`](https://github.com/xkbcommon/libxkbcommon/blob/master/include/xkbcommon/xkbcommon-keysyms.h).
     */
    xkb_keysym_t keySymbol(UInt32 keyCode);

    /**
     * @brief Keyboard state.
     *
     * XKB keyboard map state.
     */
    xkb_state *keymapState() const;

    /**
     * @brief Check the state of a modifier.
     *
     * @param name Name of the modifier "Control", "Shift", etc.
     * @returns `true` if the modifier is active and `false` otherwise.
     */
    bool isModActive(const char *name) const;

    /**
     * @brief List of pressed key codes.
     */
    const std::list<UInt32> &pressedKeys() const;

    /**
     * @brief Check if a key code is pressed.
     */
    bool isKeyCodePressed(UInt32 keyCode) const;

    /**
     * @brief Set the keyboard map.
     *
     * This method finds and assigns a keyboard map that matches the parameters passed in the arguments.\n
     * The keyboard map is automatically sent to clients when they connect to the compositor and use the [wl_keyboard](https://wayland.app/protocols/wayland#wl_keyboard) interface
     * of the Wayland protocol. If a client is already connected, the library re-sends it the new keyboard map.\n
     * The library by default assigns the "latam" layout in LCompositor::initialized().\n
     *
     * @param rules Rules on how to interpret the other arguments. Can be `nullptr`. [More information] (https://xkbcommon.org/doc/current/structxkb__rule__names.html#a0968f4602001f2306febd32c34bd2280).
     * @param model Keyboard model. Can be `nullptr`. [More information] (https://xkbcommon.org/doc/current/structxkb__rule__names.html#a0968f4602001f2306febd32c34bd2280).
     * @param layout Keyboard layouts separated by comma (e.g. "latam"). Can be `nullptr`. [More information] (https://xkbcommon.org/doc/current/structxkb__rule__names.html#a0968f4602001f2306febd32c34bd2280).
     * @param variant Variants of layouts. Can be `nullptr`. [More information] (https://xkbcommon.org/doc/current/structxkb__rule__names.html#a0968f4602001f2306febd32c34bd2280).
     * @param options Preferences not related to layouts. Can be `nullptr`. [More information] (https://xkbcommon.org/doc/current/structxkb__rule__names.html#a0968f4602001f2306febd32c34bd2280).
     */
    bool setKeymap(const char *rules = nullptr, const char *model = nullptr, const char *layout = nullptr, const char *variant = nullptr, const char *options = nullptr);

    /**
     * @brief Send a key event.
     *
     * Sends a key event to the currently focused surface.
     *
     * @param keyCode Raw key code generated by the input backend.
     * @param keyState Key state (LKeyboard::Pressed or LKeyboard::Released).
     */
    void sendKeyEvent(UInt32 keyCode, KeyState keyState);

    /**
     * @brief Send custom modifier states.
     *
     * This method allows you to send custom modifier states to the currently focused surface.
     *
     * The parameters are equivalent to those described in LKeyboard::KeyboardModifiersState.
     */
    void sendModifiersEvent(UInt32 depressed, UInt32 latched, UInt32 locked, UInt32 group);

    /**
     * @brief Send the current modifiers state.
     *
     * Sends the modifiers states to the focused surface.
     *
     * @note Louvre internally stores and updates the modifiers state, which can be accessed with modifiersState().
     */
    void sendModifiersEvent();

    /**
     * @brief Notifies a change in the modifiers state.
     *
     * Override this virtual method if you want to be notified when the modifiers state changes.
     *
     * #### Default implementation
     * @snippet LKeyboardDefault.cpp keyModifiersEvent
     */
    virtual void keyModifiersEvent(UInt32 depressed, UInt32 latched, UInt32 locked, UInt32 group);

    /**
     * @brief Notifies a key state change.
     *
     * Override this virtual method if you want to handle key press/release events.
     *
     * #### Default implementation
     * @snippet LKeyboardDefault.cpp keyEvent
     */
    virtual void keyEvent(UInt32 keyCode, KeyState keyState);

    /**
     * @brief Notifies that the focused surface changed.
     *
     * This event is triggered when the focus() property changes, either through the use of setFocus()
     * or when the current focused surface is destroyed.
     *
     * @see setFocus()
     * @see focus()
     *
     * #### Default Implementation
     * @snippet LKeyboardDefault.cpp focusChanged
     */
    virtual void focusChanged();

    LPRIVATE_IMP(LKeyboard)
};

#endif // LKEYBOARD_H
